#pragma once
#include "Function.h"
#include "Pass.hh"
#include "logging.hpp"

#include <deque>
#include <unordered_map>
#include <unordered_set>

/**
 * TODO: 完成四种前置的控制流简化，提升消除效果
 * TODO: 完成更加激进的死代码消除，消除无用的 store 指令
 * 常量传播：参见 https://www.zhihu.com/question/55976094/answer/147302764
 * 死代码消除：参见 https://www.clear.rice.edu/comp512/Lectures/10Dead-Clean-SCCP.pdf
 * 具体原理参见 编译器设计 p403
 **/
class DeadCode : public Pass {
  public:
    DeadCode(Module *m) : Pass(m) {}

    // 处理流程：两趟处理，mark 标记有用变量，sweep 删除无用指令
    void run() {
        for (auto func : m_->get_functions()) {
            init(func.get());
            mark(func.get());
            sweep(func.get());
        }

        LOG_INFO << "dead code pass erases " << ins_count << " instructions";
    }

  private:
    // 初始化，主要是 RDF 相关计算的处理（因为要用到支配树）
    void init(Function *func) {}
    void mark(Function *func) {
        work_list.clear();
        marked.clear();

        for (auto bb : func->get_basic_blocks()) {
            for (auto ins : bb->get_instructions()) {
                if (is_critical(ins)) {
                    marked[ins] = true;
                    work_list.push_back(ins);
                }
            }
        }

        while (work_list.empty() == false) {
            auto now = work_list.front();
            work_list.pop_front();

            mark(now);
        }
    }
    void mark(std::shared_ptr<Instruction> ins) {
        for (auto op : ins->get_operands()) {
            auto ptr = op.get()->shared_from_this();
            std::shared_ptr<Instruction> def = std::dynamic_pointer_cast<Instruction>(ptr);
            if (def == nullptr)
                continue;
            if (marked[def])
                continue;
            if (def->get_function() != ins->get_function())
                continue;
            marked[def] = true;
            work_list.push_back(def);
        }
    }
    bool is_instruction(std::shared_ptr<Value> val) { return std::dynamic_pointer_cast<Instruction>(val) != nullptr; }
    void sweep(Function *func) {
        for (auto bb : func->get_basic_blocks()) {
            std::list<std::shared_ptr<Instruction>> wait_del{};
            for (auto it = bb->get_instructions().begin(); it != bb->get_instructions().end();) {
                if (marked[*it]) {
                    ++it;
                    continue;
                }
                // if (is_branch(*it)) {
                //     // TODO 将 it 重写为 jump 指令
                //     // 不做这部分处理，移动到控制流相关的优化中
                // }
                // if (not(*it)->is_br()) {
                else {
                    auto tmp = *it;
                    it = bb->get_instructions().erase(it);
                    tmp->remove_use_of_ops();
                    ++ins_count;
                }
            }
        }
    }
    bool is_critical(std::shared_ptr<Instruction> ins) {
        if (ins->is_call())
            return true;
        if (ins->is_br() || ins->is_ret())
            return true;
        if (ins->is_store())
            return true;
        return false;
    }
    bool is_branch(std::shared_ptr<Instruction> ins) { return ins->is_br(); }
    // 计算出 bb 的反向支配边界，用于标记有用指令
    std::unordered_set<BasicBlock *> RDF(BasicBlock *bb) { return std::unordered_set<BasicBlock *>{}; }

    // 用以衡量死代码消除的性能
    int ins_count{0};
    std::deque<std::shared_ptr<Instruction>> work_list{};
    std::unordered_map<std::shared_ptr<Instruction>, bool> marked{};
};
