#pragma once
#include "Function.h"
#include "Pass.hh"
#include "logging.hpp"

#include <deque>
#include <unordered_map>
#include <unordered_set>

/**
 * @brief 提取基本块中的公共子表达式，全局公共子表达式消除在另一个 Pass 中实现
 * 原理参见 高级编译器设计与实现-局部公共子表达式删除
 */
class LocalCommonExpression : public Pass {
  public:
    LocalCommonExpression(Module *m) : Pass(m) {}

    void run() {
        auto func_list = m_->get_functions();
        for (auto func : func_list) {
            if (func->get_num_basic_blocks() == 0)
                continue;
            for (auto bb : func->get_basic_blocks()) {
                process_bb(bb);
            }
        }

        LOG_INFO << "local common expression process " << ins_count << " instructions";
    }

  private:
    using llu = unsigned long long;
    static constexpr llu mod = 1e9 + 7;
    // TODO: 对具有交换律的二元运算特殊处理
    // TODO: 考虑加入对 GEP 指令的特殊处理？
    class myhash {
      public:
        // TODO 经过测试发现，以下代码中的 1 具有不同的地址，较为奇怪？
        /** 表现异常的代码
            int fun() {
                int a = 1, b = 1;
                int c = a + 1;
                int d = a + 1;
                int e = c + d;
                return e;
            }
         */
        std::size_t operator()(const std::shared_ptr<BinaryInst> now) const {
            auto lhs = reinterpret_cast<uintptr_t>(now->get_operand(0).get());
            auto rhs = reinterpret_cast<uintptr_t>(now->get_operand(1).get());
            auto op = static_cast<int>(now->get_instr_type());
            // LOG_INFO << now->print() << " " << lhs << " " << rhs << " " << op;
            return std::hash<uintptr_t>()(lhs) + mod * std::hash<uintptr_t>()(rhs) + mod * mod * std::hash<int>()(op);
        };
    };

    class myequal {
      public:
        bool operator()(const std::shared_ptr<BinaryInst> &lhs, const std::shared_ptr<BinaryInst> &rhs) const {
            return lhs->get_operand(0).get() == rhs->get_operand(0).get() and
                   lhs->get_operand(1).get() == rhs->get_operand(1).get() and
                   lhs->get_instr_type() == rhs->get_instr_type();
        };
    };

    // AEB: 可用表达式的集合
    std::unordered_set<std::shared_ptr<BinaryInst>, myhash, myequal> aeb{};
    void process_bb(std::shared_ptr<BasicBlock> bb) {
        aeb.clear();

        auto inses = bb->get_instructions();
        for (auto it = inses.begin(); it != inses.end(); ++it) {
            auto ins = *it;
            if (ins->isBinary()) {
                auto bin_ins = std::dynamic_pointer_cast<BinaryInst>(ins);
                assert(bin_ins != nullptr);
                if (aeb.find(bin_ins) != aeb.end()) {
                    auto pre_ins = *aeb.find(bin_ins);
                    // 清理工作交给死代码删除
                    bin_ins->replace_all_use_with(pre_ins);
                    ++ins_count;
                } else {
                    aeb.insert(bin_ins);
                }
            }
            // 在 SSA 后旧的值就不会被修改了，因此这一步判断可以略去
            // for (auto it = aeb.begin(); it != aeb.end();) {
            //     if (changed(*it, ins))
            //         it = aeb.erase(it);
            //     else
            //         ++it;
            // }
        }
    }

    bool changed(std::shared_ptr<BinaryInst> ins, std::shared_ptr<Instruction> val) {
        if (ins->get_operand(0).get() == val.get() or ins->get_operand(1).get() == val.get())
            return true;
        return false;
    }

    int ins_count{0};
};
