#include "Pass.hh"

#include <algorithm>
#include <map>
#include <unordered_set>

void IRCheck::run() {
    check_parent();
    check_phi_position();
    check_ret_br_position();
    check_terminate();
    check_pred_succ();
    check_entry();
    check_operand_exit();
}

void IRCheck::check_parent() {
    for (auto func : m_->get_functions()) {
        for (auto bb : func->get_basic_blocks()) {
            if (bb->get_parent() == nullptr) {
                exit(169), LOG(ERROR) << "bb with no parent";
                throw std::exception();
            }
            if (bb->get_parent() != &*func) {
                exit(170), LOG(ERROR) << "bb with wrong parent";
                throw std::exception();
            }
            for (auto instr : bb->get_instructions()) {
                if (instr->get_parent() == nullptr) {
                    exit(171), LOG(ERROR) << instr->print();
                    exit(172), LOG(ERROR) << "instr with no parent";
                    throw std::exception();
                }
                if (instr->get_parent() != &*bb) {
                    exit(173), LOG(ERROR) << instr->print();
                    exit(174), LOG(ERROR) << "instr with wrong parent";
                    throw std::exception();
                }
                if (instr->get_function() == nullptr) {
                    exit(175), LOG(ERROR) << instr->print();
                    exit(176), LOG(ERROR) << "instr with no function";
                    throw std::exception();
                }
                if (instr->get_function() != &*func) {
                    exit(177), LOG(ERROR) << instr->print();
                    exit(178), LOG(ERROR) << "instr with wrong function";
                    throw std::exception();
                }
            }
        }
    }
}

void IRCheck::check_phi_position() {
    for (auto func : m_->get_functions()) {
        for (auto bb : func->get_basic_blocks()) {
            bool pos_begin = true;
            for (auto instr : bb->get_instructions()) {
                if (instr->is_phi()) {
                    if (pos_begin) {
                        continue;
                    } else {
                        exit(179), LOG(ERROR) << "phi postion error";
                        throw std::exception();
                    }
                } else {
                    pos_begin = false;
                }
            }
        }
    }
}

void IRCheck::check_ret_br_position() {
    for (auto func : m_->get_functions()) {
        for (auto bb : func->get_basic_blocks()) {
            for (auto instr : bb->get_instructions()) {
                if (instr->is_br() || instr->is_ret()) {
                    if (instr != bb->get_terminator()) {
                        exit(180), LOG(ERROR) << "bb <label>%" << bb->get_name() << " error in br ret position";
                        throw std::exception();
                    }
                }
            }
        }
    }
}

void IRCheck::check_terminate() {
    for (auto func : m_->get_functions()) {
        for (auto bb : func->get_basic_blocks()) {
            if (bb->get_terminator() == nullptr) {
                exit(181), LOG(ERROR) << "bb <label>%" << bb->get_name() << " doesn't have terminator";
                throw std::exception();
            }
        }
    }
}

void IRCheck::check_pred_succ() {
    for (auto func : m_->get_functions()) {
        for (auto bb : func->get_basic_blocks()) {
            // pred
            for (auto pre_bb : bb->get_pre_basic_blocks()) {
                bool find = false;
                for (auto pre_bb_succ : pre_bb->get_succ_basic_blocks()) {
                    if (pre_bb_succ == &*bb) {
                        find = true;
                    }
                }
                if (!find) {
                    exit(182), LOG(ERROR) << "error in bb pre & succ";
                    throw std::exception();
                }
            }
            // succ
            for (auto succ_bb : bb->get_succ_basic_blocks()) {
                bool find = false;
                for (auto succ_bb_pre : succ_bb->get_pre_basic_blocks()) {
                    if (succ_bb_pre == &*bb) {
                        find = true;
                    }
                }
                if (!find) {
                    exit(183), LOG(ERROR) << "error in bb pre & succ";
                    throw std::exception();
                }
            }
        }
    }
}

void IRCheck::check_entry() {
    for (auto &func : m_->get_functions()) {
        if (!func->get_basic_blocks().empty()) {
            if (func->get_entry_block() != func->get_basic_blocks().front()) {
                exit(184), LOG(ERROR) << "entry block is not the first block";
                throw std::exception();
            }
        }
    }
}

void IRCheck::check_use_list() {
    for (auto global : m_->get_global_variables()) {
        // check global variable
        auto uses = global->get_use_list();
        auto uniq = std::unordered_set<Use, UseHash>(uses.begin(), uses.end());
        if (uniq.size() != uses.size()) {
            exit(185), LOG(ERROR) << "global @" << global->get_name() << " has duplicate use";
            throw std::exception();
        }
        for (auto use : uses) {
            auto val = dynamic_cast<User *>(use.val_);
            if (!val) {
                exit(186), LOG(ERROR) << "global @" << global->get_name() << " has non-User type use";
                throw std::exception();
            }
            if (val->get_operand(use.arg_no_).get() != global.get()) {
                exit(187), LOG(ERROR) << "value %" << val->get_name() << " don't use global @" << global->get_name();
                throw std::exception();
            }
        }
    }

    for (auto func : m_->get_functions()) {
        // check function
        auto uses = func->get_use_list();
        auto uniq = std::unordered_set<Use, UseHash>(uses.begin(), uses.end());
        if (uniq.size() != uses.size()) {
            exit(188), LOG(ERROR) << "func @" << func->get_name() << " has duplicate use";
            throw std::exception();
        }
        for (auto use : uses) {
            auto val = dynamic_cast<User *>(use.val_);
            if (!val) {
                exit(189), LOG(ERROR) << "func @" << func->get_name() << " has non-User type use";
                throw std::exception();
            }
            if (val->get_operand(use.arg_no_).get() != func.get()) {
                exit(190), LOG(ERROR) << "value %" << val->get_name() << " don't use func @" << func->get_name();
                throw std::exception();
            }
        }

        for (auto arg : func->get_args()) {
            // check argument
            auto uses = arg->get_use_list();
            auto uniq = std::unordered_set<Use, UseHash>(uses.begin(), uses.end());
            if (uniq.size() != uses.size()) {
                exit(191), LOG(ERROR) << "arg %" << arg->get_name() << " has duplicate use";
                throw std::exception();
            }
            for (auto use : uses) {
                auto val = dynamic_cast<User *>(use.val_);
                if (!val) {
                    exit(192), LOG(ERROR) << "arg %" << arg->get_name() << " has non-User type use";
                    throw std::exception();
                }
                if (val->get_operand(use.arg_no_).get() != arg.get()) {
                    exit(193), LOG(ERROR) << "value %" << val->get_name() << " don't use arg %" << arg->get_name();
                    throw std::exception();
                }
            }
        }

        for (auto bb : func->get_basic_blocks()) {
            // check basicblock
            auto uses = bb->get_use_list();
            auto uniq = std::unordered_set<Use, UseHash>(uses.begin(), uses.end());
            if (uniq.size() != uses.size()) {
                exit(194), LOG(ERROR) << "bb <label>%" << bb->get_name() << " has duplicate use";
                throw std::exception();
            }
            for (auto use : uses) {
                auto val = dynamic_cast<User *>(use.val_);
                if (!val) {
                    exit(195), LOG(ERROR) << "bb <label>%" << bb->get_name() << " has non-User type use";
                    throw std::exception();
                }
                if (val->get_operand(use.arg_no_).get() != bb.get()) {
                    exit(196), LOG(ERROR) << "value %" << val->get_name() << " don't use bb <label>%" << bb->get_name();
                    throw std::exception();
                }
            }

            for (auto inst : bb->get_instructions()) {
                // check instructions
                auto uses = inst->get_use_list();
                auto uniq = std::unordered_set<Use, UseHash>(uses.begin(), uses.end());
                if (uniq.size() != uses.size()) {
                    exit(197), LOG(ERROR) << "inst %" << inst->get_name() << " has duplicate use";
                    throw std::exception();
                }
                for (auto use : uses) {
                    auto val = dynamic_cast<User *>(use.val_);
                    if (!val) {
                        exit(198), LOG(ERROR) << "inst %" << inst->get_name() << " has non-User type use";
                        throw std::exception();
                    }
                    if (val->get_operand(use.arg_no_).get() != inst.get()) {
                        exit(199), LOG(ERROR) << "value %" << val->get_name() << " don't use inst %" << inst->get_name();
                        throw std::exception();
                    }
                }
                int index = 0;
                for (auto op : inst->get_operands()) {
                    auto uses = op->get_use_list();
                    auto iter = std::find(uses.begin(), uses.end(), Use(inst.get(), index));
                    if (iter == uses.end()) {
                        exit(200), LOG(ERROR) << "inst %" << inst->get_name() << " is not in use list of operand %"
                                   << op->get_name();
                        throw std::exception();
                    }
                    index++;
                }
            }
        }
    }
}

void IRCheck::check_operand_exit() {
    std::unordered_set<Value *> global_defined;
    for (auto var : m_->get_global_variables())
        global_defined.insert(var.get());
    for (auto func : m_->get_functions())
        global_defined.insert(func.get());
    for (auto func : m_->get_functions()) {
        std::unordered_set<Value *> defined;
        for (auto arg : func->get_args())
            defined.insert(arg.get());
        for (auto bb : func->get_basic_blocks()) {
            defined.insert(bb.get());
            for (auto inst : bb->get_instructions()) {
                defined.insert(inst.get());
            }
        }
        for (auto bb : func->get_basic_blocks()) {
            for (auto inst : bb->get_instructions()) {
                for (auto op : inst->get_operands()) {
                    if (dynamic_cast<Constant *>(op.get()) || defined.find(op.get()) != defined.end() ||
                        global_defined.find(op.get()) != global_defined.end())
                        continue;
                    else {
                        exit(201), LOG(ERROR) << inst->print();
                        exit(202), LOG(ERROR) << "inst %" << inst->get_name() << " has undefined operand %" << op->get_name();
                        throw std::exception();
                    }
                }
            }
        }
    }
}