#include "codegen.hh"

#include <assert.h>

std::ofstream *Instgen::output{nullptr};
template <typename... Args>
void Instgen::push(Args... args) {
    (*output) << "\tpush\t"
              << "{" << (... << args) << "}\n";
}

// void Instgen::push(std::vector<Reg> args) {
//     if (args.empty())
//         return;
//     (*output) << "\tpush\t{";
//     (*output) << args.front();
//     for (auto it = std::next(args.begin()); it != args.end(); ++it) {
//         (*output) << ", " << *it;
//     }
//     (*output) << "}\n";
// }

void Instgen::push(std::vector<Reg> args) {
    for (auto arg : args)
        assert(arg.valid() and not arg.is_float);
    if (args.empty())
        return;
    (*output) << "\tpush\t{";
    (*output) << *args.begin();
    for (auto it = std::next(args.begin()); it != args.end(); ++it) {
        (*output) << ", " << *it;
    }
    (*output) << "}\n";
}

void Instgen::push_caller_saved() { (*output) << "\tpush\t{r0-r3}\n"; }

void Instgen::pop_caller_saved() { (*output) << "\tpop\t{r0-r3}\n"; }

template <typename... Args>
void Instgen::pop(Args... args) {
    (*output) << "\tpop\t"
              << "{" << (... << args) << "}\n";
}

// void Instgen::pop(std::vector<Reg> args) {
//     if (args.empty())
//         return;
//     (*output) << "\tpop\t{";
//     (*output) << args.front();
//     for (auto it = std::next(args.begin()); it != args.end(); ++it) {
//         (*output) << ", " << *it;
//     }
//     (*output) << "}\n";
// }

void Instgen::pop(std::vector<Reg> args) {
    for (auto arg : args)
        assert(arg.valid() and not arg.is_float);
    if (args.empty())
        return;
    (*output) << "\tpop\t{";
    (*output) << *args.begin();
    for (auto it = std::next(args.begin()); it != args.end(); ++it) {
        (*output) << ", " << *it;
    }
    (*output) << "}\n";
}

void Instgen::bl(std::string_view sv) { (*output) << "\tbl\t" << sv << "\n"; }

// template <typename... Args>
// void assert_regs(Args... args) {}

void Instgen::mov(Reg r1, Reg r2, std::string_view cond) {
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    if (r1 != r2)
        (*output) << "\tmov" << cond << "\t" << r1 << ", " << r2 << "\n";
}

void Instgen::mov(Reg r, int i, std::string_view cond) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmov" << cond << "\t" << r << ", #" << i << "\n";
}

void Instgen::movt(Reg r, int i, std::string_view cond) {
    assert(r.valid() and not r.is_float);
    (*output) << "\tmovt" << cond << "\t" << r << ", #" << i << "\n";
}

void Instgen::movw(Reg r, int i, std::string_view cond) {
    assert(r.valid() and not r.is_float);
    (*output) << "\tmovw" << cond << "\t" << r << ", #" << i << "\n";
}

void Instgen::sub(Reg r1, Reg r2, int n) {
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    if (n != 0)
        (*output) << "\tsub\t" << r1 << ", " << r2 << ", #" << n << "\n";
}

void Instgen::add(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tadd\t" << dest << ", " << r1 << ", " << r2 << "\n";
}
void Instgen::sub(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tsub\t" << dest << ", " << r1 << ", " << r2 << "\n";
}
void Instgen::mul(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tmul\t" << dest << ", " << r1 << ", " << r2 << "\n";
}
void Instgen::cmp(Reg r1, Reg r2) {
    assert(r1.valid() and not r1.is_float);
    assert(r1.valid() and not r2.is_float);
    (*output) << "\tcmp\t" << r1 << ", " << r2 << "\n";
}
void Instgen::tst(Reg r1, int n) { (*output) << "\ttst\t" << r1 << ", #" << n << "\n"; }

void Instgen::add(Reg dest, Reg r1, int imm) {
    assert(r1.valid() and not dest.is_float);
    assert(not r1.is_float);

    if (imm != 0)
        (*output) << "\tadd\t" << dest << ", " << r1 << ", #" << imm << "\n";
}
void Instgen::lsl(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tlsl\t" << dest << ", " << r1 << ", " << r2 << "\n";
}
void Instgen::lsr(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tlsr\t" << dest << ", " << r1 << ", " << r2 << "\n";
}
void Instgen::asr(Reg dest, Reg r1, Reg r2) {
    assert(dest.valid() and not dest.is_float);
    assert(r1.valid() and not r1.is_float);
    assert(r2.valid() and not r2.is_float);
    (*output) << "\tasr\t" << dest << ", " << r1 << ", " << r2 << "\n";
}

void Instgen::ite(std::string_view sv) { (*output) << "\tite\t" << sv << "\n"; }
void Instgen::bx(Reg r) { (*output) << "\tbx\t" << r << "\n"; }
void Instgen::b(std::string_view sv) { (*output) << "\tb\t" << sv << "\n"; }
void Instgen::beq(std::string_view sv) { (*output) << "\tbeq\t" << sv << "\n"; }
void Instgen::bne(std::string_view sv) { (*output) << "\tbne\t" << sv << "\n"; }
void Instgen::bgt(std::string_view sv) { (*output) << "\tbgt\t" << sv << "\n"; }
void Instgen::bge(std::string_view sv) { (*output) << "\tbge\t" << sv << "\n"; }
void Instgen::blt(std::string_view sv) { (*output) << "\tblt\t" << sv << "\n"; }
void Instgen::ble(std::string_view sv) { (*output) << "\tble\t" << sv << "\n"; }

void Instgen::bhi(std::string_view sv) { (*output) << "\tbhi\t" << sv << "\n"; }
void Instgen::bpl(std::string_view sv) { (*output) << "\tbpl\t" << sv << "\n"; }

void Instgen::ldr(Reg r, std::string_view sv) {
    assert(r.valid() and not r.is_float);
    (*output) << "\tldr\t" << r << ", " << sv << "\n";
}

void Instgen::movgt(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmovgt\t" << r << ", #" << i << "\n";
}
void Instgen::movlt(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmovlt\t" << r << ", #" << i << "\n";
}
void Instgen::movge(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmovge\t" << r << ", #" << i << "\n";
}
void Instgen::movle(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmovle\t" << r << ", #" << i << "\n";
}
void Instgen::movne(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmovne\t" << r << ", #" << i << "\n";
}
void Instgen::moveq(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    unsigned int ui = static_cast<unsigned int>(i);
    unsigned int nui = static_cast<unsigned int>(-i);
    assert(ui < Codegen::imm16 or nui < Codegen::imm8);
    (*output) << "\tmoveq\t" << r << ", #" << i << "\n";
}
void Instgen::movmi(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    // assert(i < Codegen::imm16 and -i < Codegen::imm8);
    (*output) << "\tmovmi\t" << r << ", #" << i << "\n";
}
void Instgen::movpl(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    // assert(i < Codegen::imm16 and -i < Codegen::imm8);
    (*output) << "\tmovpl\t" << r << ", #" << i << "\n";
}
void Instgen::movls(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    // assert(i < Codegen::imm16 and -i < Codegen::imm8);
    (*output) << "\tmovls\t" << r << ", #" << i << "\n";
}
void Instgen::movhi(Reg r, int i) {
    assert(r.valid() and not r.is_float);
    // assert(i < Codegen::imm16 and -i < Codegen::imm8);
    (*output) << "\tmovhi\t" << r << ", #" << i << "\n";
}

void Instgen::str(Reg r, Reg base, int offset, std::string_view cond) {
    assert(r.valid() and not r.is_float);
    assert(base.valid() and not base.is_float);

    if (offset != 0) {
        if (offset < 4096 and offset > -255)
            (*output) << "\tstr" << cond << "\t" << r << ", [" << base << ", #" << offset << "]\n";
        else {
            assert(r != Codegen::temp_rhs_reg and base != Codegen::temp_rhs_reg);
            Codegen::move(Codegen::temp_rhs_reg, offset, cond);
            (*output) << "\tstr" << cond << "\t" << r << ", [" << base << ", " << Codegen::temp_rhs_reg << "]\n";
        }
    } else
        (*output) << "\tstr" << cond << "\t" << r << ", [" << base << "]\n";
}

void Instgen::ldr(Reg r, Reg base, int offset, std::string_view cond) {
    assert(r.valid() and not r.is_float);
    assert(base.valid() and not base.is_float);
    if (r.is_float) {  // how do we get here?
        Instgen::vldr(r, base, offset, cond);
        return;
    }
    if (offset != 0) {
        if (offset < 4096 and offset > -255)
            (*output) << "\tldr" << cond << "\t" << r << ", [" << base << ", #" << offset << "]\n";
        else {
            assert(r != Codegen::temp_rhs_reg and base != Codegen::temp_rhs_reg);
            Codegen::move(Codegen::temp_rhs_reg, offset, cond);
            (*output) << "\tldr" << cond << "\t" << r << ", [" << base << ", " << Codegen::temp_rhs_reg << "]\n";
        }
    } else
        (*output) << "\tldr" << cond << "\t" << r << ", [" << base << "]\n";
}

void Instgen::adrl(Reg r, label l) {
    assert(r.valid() and not r.is_float);
    (*output) << "\tadrl " << r << ", " << l << "\n ";
}

void Instgen::vadd_f32(Reg r1, Reg r2, Reg r3) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    assert(r3.valid() and r3.is_float);
    (*output) << "\tvadd.f32\t" << r1 << ", " << r2 << ", " << r3 << "\n";
}
void Instgen::vsub_f32(Reg r1, Reg r2, Reg r3) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    assert(r3.valid() and r3.is_float);
    (*output) << "\tvsub.f32\t" << r1 << ", " << r2 << ", " << r3 << "\n";
}
void Instgen::vmul_f32(Reg r1, Reg r2, Reg r3) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    assert(r3.valid() and r3.is_float);
    (*output) << "\tvmul.f32\t" << r1 << ", " << r2 << ", " << r3 << "\n";
}
void Instgen::vdiv_f32(Reg r1, Reg r2, Reg r3) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    assert(r3.valid() and r3.is_float);
    (*output) << "\tvdiv.f32\t" << r1 << ", " << r2 << ", " << r3 << "\n";
}

void Instgen::vstr(Reg fr, Reg base, int offset, std::string_view cond) {
    assert(fr.valid() and fr.is_float);
    assert(base.valid() and not base.is_float);
    (*output) << "\tvstr" << cond << "\t" << fr << ", [" << base << ", #" << offset << "]\n";
}

void Instgen::vstr(Reg fr, Reg base, std::string_view cond) {
    assert(fr.valid() and fr.is_float);
    assert(base.valid() and not base.is_float);
    (*output) << "\tvstr" << cond << "\t" << fr << ", [" << base << "]\n";
}

void Instgen::vldr(Reg fr, Reg base, int offset, std::string_view cond) {
    assert(fr.valid() and fr.is_float);
    assert(base.valid() and not base.is_float);
    (*output) << "\tvldr" << cond << "\t" << fr << ", [" << base << ", #" << offset << "]\n";
}
void Instgen::vldr(Reg fr, std::string_view sv) {
    assert(fr.valid() and fr.is_float);
    (*output) << "\tvldr\t" << fr << ", " << sv << "\n";
}

void Instgen::vldr(Reg fr, Reg base, std::string_view cond) {
    assert(fr.valid() and fr.is_float);
    (*output) << "\tvldr" << cond << "\t" << fr << ", [" << base << "]\n";
}

void Instgen::vcvt_s32_f32(Reg r1, Reg r2) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    (*output) << "\tvcvt.s32.f32\t" << r1 << ", " << r2 << "\n";
}
void Instgen::vcvt_f32_s32(Reg r1, Reg r2) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    (*output) << "\tvcvt.f32.s32\t" << r1 << ", " << r2 << "\n";
}

void Instgen::vcmp_f32(Reg r1, Reg r2) {
    assert(r1.valid() and r1.is_float);
    assert(r2.valid() and r2.is_float);
    (*output) << "\tvcmp.f32\t" << r1 << ", " << r2 << "\n";
}

void Instgen::vmov(Reg r1, Reg r2, std::string_view cond) {
    assert(r1.valid());
    assert(r2.valid());
    assert(r1.is_float or r2.is_float);
    if (r1 != r2)
        (*output) << "\tvmov" << cond << "\t" << r1 << ", " << r2 << "\n";
}

void Instgen::vpush(std::vector<Reg> regs) {
    for (auto r : regs)
        assert(r.valid() and r.is_float);
    if (regs.empty())
        return;
    (*output) << "\tvpush\t{";
    (*output) << regs.front();
    for (auto it = std::next(regs.begin()); it != regs.end(); ++it)
        (*output) << ", " << *it;
    (*output) << "}\n";
}

void Instgen::vpop(std::vector<Reg> regs) {
    for (auto r : regs)
        assert(r.valid() and r.is_float);
    if (regs.empty())
        return;
    (*output) << "\tvpop\t{";
    (*output) << regs.front();
    for (auto it = std::next(regs.begin()); it != regs.end(); ++it)
        (*output) << ", " << *it;
    (*output) << "}\n";
}
