#include "AST.hpp"
#include "AlgebraicSimplification.hh"
#include "DeadCode.hh"
#include "DivisionInvariant.hh"
#include "FuncInline.hh"
#include "GepElim.hh"
#include "GlobalCommonExpression.hh"
#include "IVReduction.hh"
#include "InstructionSimplify.hh"
#include "LocalCommonExpression.hh"
#include "LoopInfo.hh"
#include "LoopInvMotion.hh"
#include "LoopSearch.hh"
#include "Mem2Reg.hh"
#include "MulWeaken.hh"
#include "Pass.hh"
#include "RemoveUselessBr.hh"
#include "SYSYCBuilder.h"
#include "SimplifyCFG.hh"
#include "SyntaxTree.hh"
#include "UselessOperationEli.hh"
#include "codegen.hh"
#include "logging.hpp"
#include "parser.hh"

#include <boost/version.hpp>
#include <cstring>
#include <fstream>
#include <iostream>
#include <string>
using std::cout;
using std::endl;
using std::string;
using std::literals::string_literals::operator""s;
using std::literals::string_view_literals::operator""sv;
SyntaxTree syntax_tree;

void print_usage() {
    // TODO
    cout << "Usage: "
         << "cc"
         << " <input_file_path>" << endl;
}

void _trigger_opt(bool &val, const char *arg, const char *argv) {
    if (strcmp(arg, argv) == 0)
        val ^= 1;  // 用以在开启 O2 时关闭特定优化
    else if (strlen(argv) > 3 and strcmp(arg, argv + 3) == 0) {
        val = false;  // 使用类似 -no-dead-code 的选项关闭特定优化
        LOG_INFO << "Turn off " << arg << " optimization";
    }
}

int main(int argc, char *argv[]) {
    bool default_value = false;
    for (int i = 1; i < argc; ++i)
        if (argv[i] == "-O2"sv) {
            exit(11);
            default_value = true;
            break;
        }

    bool silent = false;
    bool emit_llvm = false;
    bool loop_info = false;
    bool dump_graph = false;
    bool loop_sea = false;
    bool mem2reg = default_value;
    bool simplify_cfg = default_value;
    bool loop_inv = default_value;
    bool algebraic_simplification = default_value;
    bool div_inv = default_value;
    bool dead_code = default_value;
    bool gep_elim = default_value;
    bool iv_red = false;
    bool inst_simpl = default_value;
    bool lce = default_value;
    bool gce = default_value;
    bool func_inline = false;
    bool useless_eli = default_value;
    bool mul_weaken = default_value;

    string input_file_path, target_file_path;
    if (argc <= 1) {
        print_usage();
        return 1;
    }

    for (int i = 1; i < argc; ++i) {
#define trigger_opt(val, arg) _trigger_opt(val, arg, argv[i])
        trigger_opt(silent, "-S");
        trigger_opt(emit_llvm, "-emit-llvm");
        trigger_opt(mem2reg, "-mem2reg");
        trigger_opt(loop_sea, "-loop-sea");
        trigger_opt(dump_graph, "-dump-graph");
        trigger_opt(loop_inv, "-loop-inv");
        trigger_opt(algebraic_simplification, "-alge-simpl");
        trigger_opt(div_inv, "-div-inv");
        trigger_opt(simplify_cfg, "-simplify-cfg");
        trigger_opt(dead_code, "-dead-code");
        trigger_opt(loop_info, "-loop-info");
        trigger_opt(iv_red, "-iv-red");
        trigger_opt(lce, "-lce");
        trigger_opt(gce, "-gce");
        trigger_opt(func_inline, "-func-inline");
        trigger_opt(useless_eli, "-useless-eli");
        trigger_opt(mul_weaken, "-mul-weaken");
        trigger_opt(gep_elim, "-gep-elim");
        trigger_opt(inst_simpl, "-inst-simpl");
#undef trigger_opt
        if (argv[i] == "--help"sv || argv[i] == "-h"sv) {
            print_usage();
            return 0;
        } else if (argv[i] == "-o"sv) {
            if (target_file_path.empty() && i + 1 < argc) {
                target_file_path = argv[i + 1];
                i += 1;
            } else {
                print_usage();
                return 1;
            }
        } else if (argv[i][0] != '-') {
            if (input_file_path.empty()) {
                input_file_path = argv[i];
            }
        }
    }
    if (emit_llvm and not silent) {
        print_usage();
        return 1;
    }
    LOG_DEBUG << BOOST_LIB_VERSION;

    if (target_file_path.empty()) {
        auto base_start = input_file_path.rfind('/') + 1;
        if (base_start == std::string::npos)
            base_start = 0;
        auto base_end = input_file_path.rfind('.');
        auto base = input_file_path.substr(base_start, base_end);
        if (silent)
            if (emit_llvm)
                target_file_path = base + ".ll";
            else
                target_file_path = base;  // output both asm and executable + ".s";
        else
            target_file_path = "a.out";
    }
    parse_file(input_file_path);
    // syntax_tree.print();
    LOG(INFO) << "syntax tree";

    AST ast(&syntax_tree);
    auto astRoot = ast.get_root();
    LOG(INFO) << "AST";

    // ASTPrinter printer;
    // printer.visit(*astRoot);

    exit(222);

    SYSYCBuilder builder;
    builder.visit(*astRoot);

    auto module = builder.getModule();
    module->set_ptr_size((silent and not emit_llvm) ? 4 : sizeof(char *));
    auto ir = module->print();  // 为 Value 编号并打印

    PassManager pass_manager(module);
    pass_manager.add_pass<RemoveUselessBr>();
    pass_manager.add_pass<Mem2Reg>();
    if (loop_sea)
        pass_manager.pass_list.push_back({std::make_shared<loop_search>(module, dump_graph), false});

    if (loop_inv) {
        pass_manager.add_pass<LoopInvMotion>();
    }
    pass_manager.add_pass<GepElimination>();

    if (iv_red) {
        pass_manager.add_pass<IVReduction>();
    }
    if (div_inv)
        pass_manager.add_pass<DivisionInvariant>();
    if (dead_code)
        pass_manager.add_pass<DeadCode>();
    // test loop info
    if (loop_info) {
        pass_manager.add_pass<LoopInfo>();
    }
    if (algebraic_simplification) {
        pass_manager.add_pass<AlgebraicSimplification>();
    }
    if (inst_simpl) {
        pass_manager.add_pass<InstructionSimplify>();
    }
    if (dead_code)
        pass_manager.add_pass<DeadCode>();
    if (func_inline) {
        pass_manager.add_pass<FuncInline>();
        if (loop_inv) {
            pass_manager.add_pass<LoopInvMotion>();
        }
        if (iv_red) {
            pass_manager.add_pass<IVReduction>();
        }
        if (div_inv) {
            pass_manager.add_pass<DivisionInvariant>();
        }
        if (algebraic_simplification) {
            pass_manager.add_pass<AlgebraicSimplification>();
        }
        if (dead_code)
            pass_manager.add_pass<DeadCode>();
    }
    if (simplify_cfg) {
        pass_manager.add_pass<SimplifyCFG>();
    }
    if (lce)
        pass_manager.add_pass<LocalCommonExpression>();
    if (gce)
        pass_manager.add_pass<GlobalCommonExpression>();
    if (mul_weaken)
        pass_manager.add_pass<MulWeaken>();
    if (useless_eli)
        pass_manager.add_pass<UselessOperationEli>();
    if (dead_code)
        pass_manager.add_pass<DeadCode>();
    pass_manager.run();

    ir = module->print();

    ir = "target triple = \"x86_64-pc-linux-gnu\"\n\n" + ir;

    if (emit_llvm) {
        std::ofstream output_stream(target_file_path);
        output_stream << ir;
        output_stream.close();
    } else if (silent) {
        auto base = target_file_path.substr(0, target_file_path.rfind('.'));
        auto silent_path = base + ".s";
        auto ll_temp = base + ".ll";
        std::ofstream output_stream(silent_path), ll_output_stream(ll_temp);
        ll_output_stream << ir;
        ll_output_stream.close();
        Codegen cg(builder.getModule(), output_stream);
        cg.run();
        output_stream.close();

        // arm-linux-gnueabihf-gcc test.s lib/sylib.c
        // qemu-arm -L /usr/arm-linux-gnueabihf/ ./a.out
        // or in raspberry pi: gcc test.s lib/sylib.c

        std::string library_path = argv[0];
    }
    return 0;
}
